/*
 *
 * Compile:

gcc -fPIC -nostdlib -shared -o libopenfail.so \
	-Wl,--soname -Wl,libopenfail.so.0 \
	openfail.c -ldl

 * Use:

What to fail:
OPENFAIL=/path/to/fail to fail opens to files starting with the path.
OPENFAIL_EXT=.mp3 to fail opens on files ending with the extension.

How to fail:
OPENFAIL_ERRNO=N errno to fail with (default is ENOENT).
OPENFAIL_NOEXEC=1 to unset on load.
OPENFAIL_VERBOSE=1 to see verbose reporting.
OPENFAIL_SWAP=/path/to/replace path to open instead of failing.

Examples:
OPENFAIL=/dev/nst LD_PRELOAD=./libopenfail.so cat /dev/nst0
OPENFAIL_EXT=.mp3 OPENFAIL_SWAP=/tmp/rickroll.mp3 LD_PRELOAD=./libopenfail.so \
	mpg123 ~/favorite.mp3

#
# LD_PRELOAD library for capturing "open"
#
# Copyright (C) 2001-2011 Kees Cook
# kees@outflux.net, http://outflux.net/
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# http://www.gnu.org/copyleft/gpl.html

Apologies to Yggdrasil for using all their code. 
This is based on "logwrites"; also under the GNU GPL.

*/

#define _GNU_SOURCE
#include <features.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef DYNAMIC
# include <dlfcn.h>
#else
# include <syscall.h>
#endif

static int active = 0;
static int initialized = 0;
static int fail_errno = ENOENT;
static char *swap = NULL;
static char *prefix = NULL;
static int prefix_length = 0;
static char *suffix = NULL;
static int suffix_length = 0;
static char *verbose = NULL;

void initialize(void) {
	char *errno_num;

	if ( (errno_num=getenv("OPENFAIL_ERRNO")) != NULL ) {
		fail_errno = atoi(errno_num);
	}

	prefix=getenv("OPENFAIL");
	if (prefix) prefix_length = strlen(prefix);

	suffix=getenv("OPENFAIL_EXT");
	if (suffix) suffix_length = strlen(suffix);

	swap = getenv("OPENFAIL_SWAP");

	verbose = getenv("OPENFAIL_VERBOSE");

	if (prefix || suffix) {
		active=1;
	}

       	if (getenv("OPENFAIL_NOEXEC")) {
		// stop ourselves from going over exec
		unsetenv("OPENFAIL");
		unsetenv("OPENFAIL_EXT");
	}

	initialized=1;
}

#ifdef DYNAMIC

static int (*old_open)(const char *path, int flags, int mode);
static int (*old___open)(const char *path, int flags, int mode);


int
_init(void) {
        old_open        = dlsym(RTLD_NEXT, "open");
        old___open      = dlsym(RTLD_NEXT, "__open");

	initialize();

	return 0;
}

#else

static int old_open(const char *path, int flags, int mode) {
        return syscall(SYS_open, path, flags, mode);
}
#define old___open old_open

#endif



int
myopen ( const char *filename, int flags, int mode ) {
    int result;

    if (active) {
	int triggered = 0;

        if (verbose) {
            printf("Checking: '%s'\n",filename);
        }

	if (!triggered && prefix) {
	        if (strncmp(filename, prefix, prefix_length)==0) {
			triggered = 1;
	        }
	}
	if (!triggered && suffix) {
		char *match;

		match = strstr(filename, suffix);
		if (match && match[suffix_length] == '\0') {
			triggered = 1;
		}
	}

	if (triggered) {
		if (swap) {
			filename = swap;
		}
		else {
			errno=fail_errno;
			return -1;
		}
	}
    }

    result = old_open(filename, flags, mode);
    return result;
}

int
open ( const char *filename, int flags, int mode ) {
    if (!initialized) initialize();

    return myopen(filename, flags, mode );
}

int
open64 ( const char *filename, int flags, int mode ) {
    if (!initialized) initialize();

    return myopen(filename, flags, mode );
}

#if defined(__GLIBC__) && (((__GLIBC__ == 2) && (__GLIBC_MINOR__ >= 1)) || __GLIBC__ > 2)
int
__open ( const char *filename, int flags, int mode ) {
    if (!initialized) initialize();

    return myopen(filename, flags, mode );
}

int
__open64 ( const char *filename, int flags, int mode ) {
    if (!initialized) initialize();

    return myopen(filename, flags, mode );
}
#endif
